@@ -1,11 +1,10 @@
-
 use 5.008;
 use strict;
 use warnings FATAL => 'all';
 
 package namespace::local;
 
-our $VERSION = '0.05';
+our $VERSION = '0.06';
 
 =head1 NAME
 
@@ -75,6 +74,30 @@ This emulates L<namespace::clean>, by which this module is clearly inspired.
 
     # now define public functions here
 
+=head1 OPTIONS
+
+Extra options may be passed to namespace::local:
+
+=head2 -target => Package::Name
+
+Act on another package instead of the caller.
+Note that L<namespace::local> is only meant to be used in BEGIN phase.
+
+=head2 -except => \@list
+
+Exempt symbols mentioned in list (with sigils)
+from the module's action.
+
+No sigil means a function.
+Only names made of word characters are supported.
+
+=head2 -except => <regex>
+
+Exempt symbols with names matching the regular expression
+from the module's action.
+
+Note that sigils are ignored here.
+
 =head1 EXEMPTIONS
 
 The following symbols are not touched by this module, to avoid breaking things:
@@ -139,33 +162,29 @@ use Carp;
 use B::Hooks::EndOfScope 'on_scope_end';
 use Scalar::Util qw( weaken );
 
-my %known_args;
-$known_args{$_}++ for qw(-above -below -around);
-
 my $last_command;
 
 sub import {
-    my ($class, $action) = @_;
+    my $class = shift;
 
-    # TODO more options (-except, -only, -target etc)
-    $action ||= '-around';
-    croak "Unknown argument $action"
-        unless $known_args{$action};
+    my $command = namespace::local::_command->new( target => scalar caller );
+    $command->parse_options( @_ );
 
-    # multiple callback may interfere, so accumulate
-    # ALL events at current scope end and make them play along well
-    my $command = namespace::local::_command->new;
+    # on_scope_end callback execution order is direct
+    # we need reversed order, so use a stack of commands.
     $last_command->set_next( $command ) if $last_command;
     $last_command = $command;
     weaken $last_command; # to avoid leaks
 
-    $command->prepare( action => $action, target => scalar caller );
+    $command->prepare;
 
     on_scope_end {
         $command->execute;
     };
 };
 
+;
+
 # Hide internal OO engine
 # Maybe it will be released later...
 
@@ -175,26 +194,83 @@ package
 use Carp; # too
 our @CARP_NOT = qw(namespace::local);
 
+# target => package_name is required
 sub new {
-    return bless { }, shift;
+    my ($class, %opt) = @_;
+
+    # TODO check options
+    $opt{except_rex} = qr/^[0-9]+$|^-$/; # no matter what, exempt $_, $1, ...
+    $opt{action} = '-around';
+
+    # Skip some well-known variables and functions
+    # Format: touch_not{ $name }{ $type }
+    # NOTE if you change the list, also change the EXEMPTIONS section in the POD.
+    $opt{touch_not}{$_}{ARRAY}++  for qw( CARP_NOT EXPORT EXPORT_OK ISA );
+    $opt{touch_not}{$_}{CODE}++   for qw( AUTOLOAD DESTROY import );
+    $opt{touch_not}{$_}{IO}++     for qw( DATA STDERR STDIN STDOUT );
+    $opt{touch_not}{$_}{SCALAR}++ for qw( AUTOLOAD a b );
+
+    return bless \%opt, $class;
 };
 
 sub set_next {
     my ($self, $next) = @_;
-    carp "Next already set"
-        if $self->{next};
     $self->{next} = $next;
 };
 
+my %known_action;
+$known_action{$_}++ for qw(-above -below -around);
+
+# this changes nothing except the object itself
+sub parse_options {
+    my $self = shift;
+
+    while (@_) {
+        my $arg = shift;
+        if ( $known_action{$arg} ) {
+            $self->{action} = $arg;
+        } elsif ($arg eq '-target') {
+            $self->{target} = shift;
+        } elsif ($arg eq '-except') {
+            my $cond = shift;
+            if (ref $cond eq 'Regexp') {
+                $self->{except_rex} = qr((?:$self->{except_rex})|(?:$cond));
+            } elsif (ref $cond eq 'ARRAY') {
+                $self->touch_not( @$cond );
+            } else {
+                _croak( "-except argument must be regexp or array" )
+            };
+        } else {
+            _croak( "unknown option $arg" );
+        };
+    };
+};
+
+# TODO join with @TYPE array from below
+my %sigil = (
+    ''  => 'CODE',
+    '$' => 'SCALAR',
+    '%' => 'HASH',
+    '@' => 'ARRAY',
+);
+
+sub touch_not {
+    my ($self, @list) = @_;
+
+    foreach (@list) {
+        /^([\$\@\%]?)(\w+)$/
+            or _croak( "cannot exempt sybmol $_: unsupported format" );
+        $self->{touch_not}{$2}{ $sigil{$1} }++
+    };
+};
+
+# side effects + setup self->execute
 sub prepare {
-    my ($self, %opt) = @_;
+    my $self = shift;
 
-    my $target = $opt{target};
-    my $action = $opt{action};
+    my $action = $self->{action};
 
-    my $control = namespace::local::_izer->new( target => $target );
-
-    $control->save_all;
+    my $table = $self->read_symbols;
 
     # FIXME UGLY HACK
     # Immediate backup-and-restore of symbol table
@@ -203,92 +279,33 @@ sub prepare {
     #     thus preventing subsequent imports from leaking upwards
     # I do not know why it works, it shouldn't.
     if ($action eq '-around') {
-        $control->restore_all;
+        $self->erase_symbols;
+        $self->write_symbols( $table );
     };
 
     if ($action eq '-above' ) {
-        $self->{action} = sub {
-            $control->erase_known;
+        $self->{todo} = sub {
+            $self->erase_only_symbols( $table );
         };
     } else {
-        $self->{action} = sub {
-            $control->restore_all;
+        $self->{todo} = sub {
+            $self->erase_symbols;
+            $self->write_symbols( $table );
         };
     };
 };
 
-# dumb execution, for now
 sub execute {
     my ($self) = @_;
 
-    # always execute stacked command in reverse order
+    # always execute stacked commands in reverse order
     $self->{next}->execute
         if $self->{next};
 
-    $self->{action}->()
+    $self->{todo}->()
         unless $self->{done}++;
 };
 
-package
-    namespace::local::_izer;
-
-use Carp; # too
-our @CARP_NOT = qw(namespace::local);
-
-# TODO use Package::Stash?
-sub new {
-    my ($class, %opt) = @_;
-
-    # TODO validate options better
-    # Assume caller?
-    croak "target package not specified"
-        unless defined $opt{target};
-
-    return bless {
-        target  => $opt{target},
-        names   => [],
-        content => {},
-    }, $class;
-
-    # TODO read names here?
-};
-
-sub save_all {
-    my $self = shift;
-
-    my @names = $self->read_names;
-
-    # Shallow copy of symbol table does not work for all cases,
-    #     or it would've been just '%content = %{ $target."::" }
-    $self->save_globs( @names );
-
-    $self->{names} = \@names;
-    return $self;
-};
-
-sub restore_all {
-    my $self = shift;
-
-    # Erase _all_ globs, then restore those known to us
-    $self->erase_unknown;
-    $self->restore_globs( @{ $self->{names} } );
-
-    return $self;
-};
-
-sub erase_known {
-    my $self = shift;
-
-    $self->erase_globs( @{ $self->{names} } );
-};
-
-sub erase_unknown {
-    my $self = shift;
-
-    my @todo = grep { !exists $self->{content}{$_} } $self->read_names;
-    $self->erase_globs( @todo );
-};
-
 # in: package name
 # out: sorted & filtered list of symbols
 
@@ -301,9 +318,10 @@ sub read_names {
     my $self = shift;
 
     my $package = $self->{target};
+    my $except = $self->{except_rex};
 
     my @list = sort grep {
-        /^\w+$/ and !/^[0-9]+$/ and $_ ne '_'
+        /^\w+$/ and $_ !~ $except
     } do {
         no strict 'refs'; ## no critic
         keys %{ $package."::" };
@@ -315,59 +333,79 @@ sub read_names {
 # Don't touch NAME, PACKAGE, and GLOB itself
 my @TYPES = qw(SCALAR ARRAY HASH CODE IO FORMAT);
 
-# Skip some well-known variables and functions
-# Format: touch_not{ $name }{ $type }
-# NOTE if you change the list, also change the EXEMPTIONS section in the POD.
-my %touch_not;
-$touch_not{$_}{ARRAY}++  for qw( CARP_NOT EXPORT EXPORT_OK ISA );
-$touch_not{$_}{CODE}++   for qw( AUTOLOAD DESTROY import );
-$touch_not{$_}{IO}++     for qw( DATA STDERR STDIN STDOUT );
-$touch_not{$_}{SCALAR}++ for qw( AUTOLOAD a b );
-
 # In: package
 # Out: (none)
 # Side effect: destroys symbol table
-sub erase_globs {
-    my ($self, @names) = @_;
-    my $package = $self->{target};
+sub erase_symbols {
+    my ($self, $list) = @_;
 
-    foreach my $name( @names ) {
-        next if $touch_not{$name};
+    my $package = $self->{target};
+    $list ||= [ $self->read_names ];
+
+    foreach my $name( @$list ) {
+        next if $self->{touch_not}{$name};
         no strict 'refs'; ## no critic
         delete ${ $package."::" }{$name};
     };
 };
 
+sub erase_only_symbols {
+    my ($self, $table) = @_;
+
+    my $package = $self->{target};
+    my @list = keys %$table;
+
+    # read all necessary symbols
+    my $current = $self->read_symbols( \@list );
+
+    foreach my $name ( @list ) {
+        $table->{$name}{$_} and delete $current->{$name}{$_}
+            for @TYPES;
+    };
+
+    $self->erase_symbols( \@list );
+    $self->write_symbols( $current );
+};
 
 # In: package, symbol
 # Out: a hash with glob content
-sub save_globs {
-    my ($self, @names) = @_;
+sub read_symbols {
+    my ($self, $list) = @_;
 
     my $package = $self->{target};
+    $list ||= [ $self->read_names ];
 
-    foreach my $name ( @names ) {
+    my %content;
+    foreach my $name ( @$list ) {
         foreach my $type (@TYPES) {
             my $value = do {
                 no strict 'refs'; ## no critic
                 *{$package."::".$name}{$type};
             };
-            $self->{content}{$name}{$type} = $value if defined $value;
+            $content{$name}{$type} = $value if defined $value;
         };
     };
+
+    return \%content;
 };
 
 # In: package, symbol, hash
 # Out: (none)
 # Side effect: recreates *package::symbol
-sub restore_globs {
-    my ($self, @names) = @_;
+sub write_symbols {
+    my ($self, $table) = @_;
 
     my $package = $self->{target};
 
-    foreach my $name( @names ) {
-        my $copy = $self->{content}{$name};
-        if ( my $skip = $touch_not{$name} ) {
+    # TODO lumping touch_not and $table together is a workaround
+    #     for erase_symbols skipping over touch_not altogether.
+    # It does not belong here, and should be fixed at some point.
+    my %uniq;
+    $uniq{$_}++ for keys %$table, keys %{ $self->{touch_not} };
+
+    foreach my $name( keys %uniq ) {
+        my $copy = $table->{$name} || {};
+        if ( my $skip = $self->{touch_not}{$name} ) {
             foreach my $type (keys %$skip) {
                 my $value = do {
                     no strict 'refs'; ## no critic
@@ -390,6 +428,10 @@ sub restore_globs {
     };
 };
 
+sub _croak {
+    croak ("namespace::local: ".shift);
+};
+
 =head1 AUTHOR
 
 Konstantin S. Uvarin, C<< <khedin at gmail.com> >>
